----------------------------------------------------------------------------------
--
--	Address decoder for fast 65C02 with local RAM.
--
-- This should be the only place where host-specific information resides.
-- Knows which addresses are to be cached internally vs. which ones are external.
-- Knows which address to put onto the bus while the CPU is busy internally.
-- Memory map to apply is selected via switch inputs.
--
--	juergen@e-basteln.de
--
--	jm 14.4.2020	original version, some dummy address models
-- jm 19.4.2020	address models for Mephisto, Apple II
-- jm  3.5.2020	Apple II $800..BFF declared internal (no text/low res page 2, fast Applesoft)
-- jm  4.5.2020	Commodore PET (up to 8032) added
-- jm  8.5.2020	host-specific idle page address added
-- jm 24.5.2020	idle address for Mephisto set to $FF00 (on-board ROM)
-- jm 28.5.2020	experimental version - alternate memory maps for Mephisto, full 16 bit address to select
-- jm 30.5.2020	memory map for Polgar added; new idle address to avoid Polgar conflict
-- jm 10.6.2020	fixed bug in $cxxx 16-bit address range for Apple II map
-- jm  2.7.2020	updated memory maps for various Mephisto systems
-- jm 10.7.2020	detect "slow" operation, started by IO access, terminated by RTS/RTI
--	jm 30.7.2020	refactored -- RTS/JSR detection moved to main unit's FSM
-- jm  7.8.2020	narrower "slow" address ranges for Milano, Apple II
-- jm 10.8.2020	add Mephisto MM II, Novag Constellation Forte, Mephisto Rebell 5.0, Chafitz ARB & MGS (untested)
-- jm 11.8.2020	"cached" video RAM for Apple II and CBM 8032: only write access is external
-- jm 12.8.2020	Fidelity Elite A/S (untested)
-- jm 14.8.2020	Novag Super Constellation (untested)
-- jm 15.8.2020	Novag Super Expert C (untested)
-- jm 17.9.2020	MM II now with slow-mode access for all I/O
-- jm 18.9.2020	Chafitz ARB now with slow-mode access for all I/O
--
-- to do:
-- accelerate!
-- Novag Super Expert C - test
-- / Novag Super Constellation
-- / Apple with/without video RAM caching, for comparison
--
---------------------------------------------------------------------------------- 
-- Switch settings: See constant definitions below
-- All other maps default to "all external addresses", i.e. no acceleration. 
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

entity decoder is
	port (
		clk			: in	std_logic;									-- cpu clock
		switch		: in	std_logic_vector (1 to 4);				-- select memory model; numbered like the physical DIP switch
		adr			: in	std_logic_vector (15 downto 0);		-- current memory address
		data			: in	std_logic_vector (7 downto 0);		-- data bus; req'd to track bank switching in some hosts
		write			: in 	std_logic;									-- is this a write cycle?
		
		is_ext		: out	std_logic;									-- current address is external (not fast RAM/ROM)
		is_slow		: out	std_logic;									-- last clk's address triggers slow (time critical) mode 
		idle			: out	std_logic_vector (7 downto 0);		-- memory page to read during "idle" cycles (CPU works internally)
		
		init_adr		: out std_logic_vector (15 downto 0);		-- address for initial write (host initialization)
		init_data	: out std_logic_vector (7 downto 0)			-- data for initial write (host initialization)
	);
end decoder;

architecture Behavioral of decoder is

	subtype dips is std_logic_vector (1 to 4);

	constant dip_mephisto_mm2		: dips := "1000";				-- Mephisto Modular II & B&P
	constant dip_mephisto_mm45		: dips := "0100";				-- Mephisto Modular IV & V, Rebel Portoroz
	constant dip_mephisto_polgar	: dips := "1100";				-- Mephisto Polgar
	constant dip_mephisto_milano	: dips := "0010";				-- Mephisto Milano & Nigel Short
	constant dip_mephisto_rebel5	: dips := "1010";				-- Mephisto Rebell 5.0
	constant dip_novag_con_forte	: dips := "0110";				-- Novag Constellation Forte A/B
	constant dip_chafitz_arb		: dips := "1110";				-- Chafitz ARB & MGS
	constant dip_apple_ii			: dips := "0001";				-- Apple II, video RAM cached in BRAM
	constant dip_pet					: dips := "1001";				-- Commodore PET/CBM up to 8032
	constant dip_fidel_elite		: dips := "0101";				-- Fidelity Elite A/S (Glasgow, Budapest); Prestige
	constant dip_novag_supercon	: dips := "1101";				-- Novag Super Constellation
	constant dip_novag_superexp	: dips := "0011";				-- Novag Super Expert C -- not implemented yet ###
	constant dip_apple_ii_nocache	: dips := "1011";				-- Apple II -- remove after testing cache performance ###
	constant dip_avail0				: dips := "0111";
	constant dip_avail1				: dips := "1111";
	
	signal novag_bank : std_logic := '1';							-- active ROM bank for Novag Super Expert C
	
begin

	---------------------------------------------- 
	-- Set address and data for host initialization prior to RAM/ROM download.
	-- Required for some bank switching schemes (guarantee defined state)
	---------------------------------------------- 

	process (clk)
	begin
		if rising_edge(clk) then
			case switch is

--				when dip_novag_superexp =>							-- Novag Super Expert C
--					init_adr <= x"1ff5";
--					init_data <= x"00";
					
				when others =>
					init_adr <= x"0000";
					init_data <= x"00";
			
			end case;
		end if;
	end process;

	---------------------------------------------- 
	-- Detect start of "slow mode".
	-- Synchronous to relax timing; 
	-- can only occur during external bus cycles (as only I/O addresses can trigger slow mode).
	-- Triggered by access to I/O addresses which are known to be timing-critical.
	-- Status tracking and RTI/RTS detection in main FSM outside of this unit.
	---------------------------------------------- 

	process (clk)
	begin
		if rising_edge(clk) then
			case switch is
			
				when dip_mephisto_mm2 =>							-- Mephisto MM II & B&P
					if adr (15 downto 12) = x"3" 
						or adr (15 downto 12) = x"2"
						or adr (15 downto 12) = x"1"
					then is_slow <= '1';
					else is_slow <= '0';
					end if;

				when dip_mephisto_rebel5 =>						-- Mephisto Rebell 5.0
					if adr (15) = '0' and adr (14 downto 13) /= "00" -- $2000..$7FFF 
--					if adr (15 downto 12) = x"2" 					-- speaker at $2006
--						or adr (15 downto 12) = x"5" 				-- LCD at $5000?
					then is_slow <= '1';
					else is_slow <= '0';
					end if;

				when dip_mephisto_mm45 =>							-- Mephisto MM IV, MM V
					if adr = x"3406"									-- speaker
					then is_slow <= '1'; 
					else is_slow <= '0';
					end if;	

				when dip_mephisto_polgar =>						-- Mephisto Polgar 
					if adr = x"2004"		 							-- speaker
					then is_slow <= '1'; 
					else is_slow <= '0';
					end if;	

				when dip_mephisto_milano =>						-- Mephisto Milano & Nigel Short
--					if adr (15 downto 6) = "0001111111" then	-- any I/O address
					if adr = x"1FF0" 									-- speaker
					then is_slow <= '1'; 
					else is_slow <= '0';
					end if;	

				when dip_novag_con_forte =>						-- Novag Constellation Forte A/B
					if adr (15 downto 10) = "000111" 			-- I/O from $1C00 to $1FFF
					then is_slow <= '1';
					else is_slow <= '0';
					end if;
				
				when dip_chafitz_arb =>								-- Chafitz ARB (Sargon 2.5, Sargon 4.0) & MGS
					if adr (15 downto 13) = "100" 				-- I/O (VIA) mirrored from $8000..$9FFF
					then is_slow <= '1';
					else is_slow <= '0';
					end if;
--					is_slow <= '0';									-- VIA 6522, no time-critical access identified 

				when dip_fidel_elite =>								-- Fidelity Elite A/S (Glasgow, Budapest); Prestige
					if adr (15 downto 12) = "0111"				-- all I/O slow? at $7xxx
					then is_slow <= '1';
					else is_slow <= '0';
					end if;

				when dip_novag_supercon	=>							-- Novag Super Constellation
					if adr (15 downto 8) = x"1e"					-- $1Exx
					then is_slow <= '1';
					else is_slow <= '0';
					end if;
					
--				when dip_novag_superexp =>							-- Novag Super Expert C
--					if adr (15 downto 4) = x"1ff"					-- $1FFx
--					then is_slow <= '1';
--					else is_slow <= '0';
--					end if;	
--					if (adr = x"1ff5") and (write = '1') then	-- ROM bank switching
--						novag_bank <= not data (0);				-- track status
--					end if;					

--				when dip_apple_ii | dip_apple_II_nocache =>	-- Apple II, II plus
--					if adr (15 downto 8) = x"C0" 					-- I/O address range
--						and (adr(7) = '1'	 							-- slot I/O
--						or adr(5) = '1')								-- $C02x cassette out, $C03x speaker, $C06x cassette in, $C07x paddles
----					if adr (15 downto 4) = x"C03" 				-- speaker
----						or adr (15 downto 4) = x"C07" 			-- game controller
----						or adr (15 downto 4) = x"C02" 			-- cassette out
----						or adr (15 downto 4) = x"C06" 			-- cassette in et al.
----						or adr (15 downto 7) = "110000001" 		-- slot I/O
--					then is_slow <= '1'; 
--					else is_slow <= '0';
--					end if;	
--				
--				when dip_pet =>										-- Commodore PET up to CBM 8032
--					if adr (15 downto 8) = x"E8"		 			-- any PIA/VIA access -- may be too much? ###
--					then is_slow <= '1'; 
--					else is_slow <= '0';
--					end if;	

				when others	=>
					is_slow <= '0';

			end case;		
		end if;	
	end process;

	---------------------------------------------- 
	-- Detect external vs. internal access (async.)
	---------------------------------------------- 

	process (adr, switch, write, novag_bank)
	begin
	
		case switch is
		
			when dip_mephisto_mm2 =>						-- Mephisto MM II
				if adr (15 downto 12) = "0001"			-- $1000..$1FFF 
					or adr (15 downto 13) = "001"			-- $2000..$3FFF
				then is_ext <= '1';
				else is_ext <= '0';
				end if;

			when dip_mephisto_rebel5 =>					-- Mephisto Rebell 5.0
				if adr (15 downto 13) = "001" 			-- I/O from $2000..$3FFF
					or adr (15 downto 14) = "01"			-- $4000..$7FFF
				then is_ext <= '1';
				else is_ext <= '0';
				end if;

			when dip_mephisto_polgar | dip_mephisto_mm45 =>		-- Mephisto Polgar, MM IV, MM V
				if adr (15 downto 13) = "001" 			-- I/O from $2000..$3FFF
				then is_ext <= '1';
				else is_ext <= '0';
				end if;
				
			when dip_mephisto_milano =>					-- Mephisto Milano & Nigel Short
				if adr (15 downto 6) = "0001111111"		-- I/O from $1FC0..$1FFF 
				then is_ext <= '1';
				else is_ext <= '0';
				end if;

			when dip_novag_con_forte =>					-- Novag Constellation Forte A/B
				if adr (15 downto 10) = "000111" 		-- I/O from $1C00 to $1FFF
				then is_ext <= '1';
				else is_ext <= '0';
				end if;
				
			when dip_chafitz_arb =>							-- Chafitz ARB (Sargon 2.5, Sargon 4.0) and MGS
				if adr (15 downto 13) = "100" 			-- I/O mirrored from $8000..$9FFF
				then is_ext <= '1';
				else is_ext <= '0';
				end if;
				
			when dip_fidel_elite =>							-- Fidelity Elite A/S (Glasgow, Budapest); Prestige
				if adr (15 downto 12) = "0111"			-- I/O at $7xxx
				then is_ext <= '1';
				else is_ext <= '0';
				end if;
				
			when dip_novag_supercon	=>						-- Novag Super Constellation
				if adr (15 downto 12) = "0001"			-- I/O at $1xxx
				then is_ext <= '1';
				else is_ext <= '0';
				end if;

--			when dip_novag_superexp =>							-- Novag Super Expert C
--				if adr (15 downto 4) = x"1ff"					-- I/O at $1FFx
--					or (adr(15)='1' and novag_bank='0')		-- ROM bank 0 (opening book?) stays external ###
--				then is_ext <= '1';
--				else is_ext <= '0';
--				end if;		
					
--			when dip_apple_II_nocache =>					-- Apple II, GR2 mapped internally (fast Applesoft), no video RAM caching
--				if adr (15 downto 10) = "000001"  		-- $400..$7FF, primary text/low-res graphics
----				or adr (15 downto 10) = "000010" 	-- $800..$BFF, Applesoft prog/variables; also GR2
--					or adr (15 downto 13) = "001" 		-- $2000..$3FFF high-res graphics
--					or adr (15 downto 13) = "010" 		-- $4000..$5FFF high-res graphics
--					or adr(15 downto 12) = x"c"			-- memory-mapped I/O and slot card ROM
--					or adr = x"e000"							-- workaround: don't let the DOS language card reset mess with the ROM ###
--				then is_ext <= '1';
--				else is_ext <= '0';
--				end if;
--				
--			when dip_apple_II =>								-- Apple II, GR2 mapped internally (fast Applesoft), video RAM cached in BRAM
--				if ((write = '1') and (
--						adr (15 downto 10) = "000001"  	-- $400..$7FF, primary text/low-res graphics
----						or adr (15 downto 10) = "000010" -- $800..$BFF, Applesoft prog/variables; also GR2
--						or adr (15 downto 13) = "001" 	-- $2000..$3FFF high-res graphics
--						or adr (15 downto 13) = "010" 	-- $4000..$5FFF high-res graphics
--					)) 
--					or adr(15 downto 12) = x"c"			-- memory-mapped I/O and slot card ROM
--					or adr = x"e000"							-- workaround: don't let the DOS language card reset mess with the ROM ###
--				then is_ext <= '1';
--				else is_ext <= '0';
--				end if;
--				
--			when dip_pet =>									-- Commodore PET, up to CBM 8032; only I/O mapped externally
--				if (write = '1' and adr (15 downto 12) = x"8") 	-- video RAM
--					or adr(15 downto 8) = x"e8"			-- memory mapped I/O (PIA, VIA)
--				then is_ext <= '1';
--				else is_ext <= '0';
--				end if;
								
			when others	=> 									-- all access is external, no accelaration
				is_ext <= '1';

		end case;
	end process;

	idle 		<=	x"f8";		-- should be harmless in all hosts
	
end Behavioral;

